/*  Copyright (C) 2014 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file Annotation.hxx
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 19.05.2014
 *  Declaraion of parameter class Annotation.
 */

#ifndef _ANNOTATION_HXX_
#define _ANNOTATION_HXX_

#include <algorithm>

//#include <CImg.h>
#include <stdexcept>
#include <iostream>
#include <iomanip>
#include <list>

namespace annotation {

	/// Annotation class which keeps all common parameters for sub annotation classes
	class Annotation;

	/// Annotation box element keeps two coordinates for upper left and lower right points
	class AnnotationPoints;

	/// Frame objects keep id, filename and pointers to all corresponding annotation objects
	class Frame;


	/// Keeps all Sequence objects
	class AnnotationList {

	public:		
		/// add frame if it does not exist already
		void addFrame( Frame _frame );

		/// add annotation
		int addAnnotationPoints(AnnotationPoints _annotation);

		/// return pointer to according frame
		const Frame* getFrame( std::string _filename ) const;

		/// return pointer to according annotation
		const AnnotationPoints* getAnnotationPoints(int _annotationId) const;

		/// get annotations list of according Frame
		const std::vector< const AnnotationPoints* > getAnnotationPoints(std::string _filename);

		/// return vector of all frames
		const std::vector< const Frame* > getFrames() const;

		/// delete annotation and annotationId in corresponding frame object
		bool deleteAnnotation( int _annotationId );

	private:
		/// A map which maps filenames to frame objects
		std::map< std::string, Frame > frames;

		/// A map which maps annatation ids to annotation box objects
		std::map< int, AnnotationPoints > annotationPoints;

	};

	/// Annotation class which keeps all common parameters for sub annotation classes
	class Annotation
	{
	public:

		/// Id of annotation object, which is unique for a specific frame
		int annotationId ;

		/// Frame filename
		std::string filename;

		/// flag for internally deleted annotations
		bool deleted;

		/// flag for finalized annotations
		bool finalized;

		/// state of task result
		std::string taskResultState;

		/// task result id
		std::string taskResultId;


		/// default constructor
		Annotation(std::string _filename, int _annotationId = -1, bool _deleted = false, bool _finalized = false, std::string _taskResultId = "", std::string _taskResultState = "") :
			filename(_filename), annotationId(_annotationId), deleted(_deleted), finalized(_finalized), taskResultId(_taskResultId), taskResultState(_taskResultState)
		{
		}

		virtual ~Annotation() {} ;

	} ;


	/// class for describing an annotation element (box, contour...)
	class AnnotationPoints : public Annotation
	{
	public:

		/// list of points
		/** 
			List of points. If the annotation object is a box annotation, this list contains two points.
			For contour annotations, the list contains an arbitrary number of points.
		*/
		std::vector< std::pair<float, float> > points;

		/// default constructor
		AnnotationPoints(std::string _filename, std::vector< std::pair<float, float> > _points) :
			Annotation(_filename), points(_points)
		{}

		// constructor if annotation id is known
		AnnotationPoints(std::string _filename, int _annotationId, std::vector< std::pair<float, float> > _points) :
			Annotation(_filename, _annotationId), points(_points)
		{}

		// constructor if annotation id and state (deleted, final) is known
		AnnotationPoints(std::string _filename, bool _deleted, bool _finalized, std::vector< std::pair<float, float> > _points) :
			Annotation(_filename, -1, _deleted, _finalized), points(_points)
		{}

		// constructor if annotation id and state (deleted, final) is known plus task result id
		AnnotationPoints(std::string _filename, bool _deleted, bool _finalized, std::string _taskResultId, std::vector< std::pair<float, float> > _points) :
			Annotation(_filename, -1, _deleted, _finalized, _taskResultId), points(_points)
		{}

		// constructor if state (deleted, final)
		AnnotationPoints(std::string _filename, int _annotationId, bool _deleted, bool _finalized, std::vector< std::pair<float, float> > _points) :
			Annotation(_filename, _annotationId, _deleted, _finalized), points(_points)
		{}

		// constructor with all parameters known
		AnnotationPoints(std::string _filename, int _annotationId, bool _deleted, bool _finalized, std::string _taskResultId, std::vector< std::pair<float, float> > _points) :
			Annotation(_filename, _annotationId, _deleted, _finalized, _taskResultId), points(_points)
		{}

		// constructor with all parameters known
		AnnotationPoints(std::string _filename, int _annotationId, bool _deleted, bool _finalized, std::string _taskResultId, std::string _taskResultState, std::vector< std::pair<float, float> > _points) :
			Annotation(_filename, _annotationId, _deleted, _finalized, _taskResultId, _taskResultState), points(_points)
		{}

		virtual ~AnnotationPoints() {};

	} ;

	
	/// Frame objects keep id, filename and pointers to all corresponding annotation objects
	class Frame{
	public:
		/// Frame id
		int frameId;

		/// Frame filename
		std::string filename;

		/// pointers to all corresponding annotation elements
		std::vector< int > annotationIds;

		/// return pointer to frame with frameId
		//Annotation* operator()(size_t annotationId);
		
		Frame ( std::string _filename, int _frameId = -1 ) : 
			frameId(_frameId), filename(_filename){}

		//Frame ( int _frameId, std::string _filename, std::vector< int > &_annotationIds ) : 
		//	frameId(_frameId), filename(_filename), annotationIds(_annotationIds){}


	};


	/// add frame if it does not exist already
	void AnnotationList::addFrame( Frame _frame ){
		frames.insert( std::pair< std::string, Frame >( _frame.filename, _frame ) );
	}


	/// add annotation
	int AnnotationList::addAnnotationPoints(AnnotationPoints _annotation){

		// check if corresponding frame exists
		std::map<std::string,Frame>::iterator it = frames.find( _annotation.filename );

		if (it == frames.end()){
			Frame frame(_annotation.filename);
			addFrame(frame);
			it = frames.find(_annotation.filename);
		}

		int newId = 0;

		if (it != frames.end()){
			
			// if _annotation.Id == -1, apply new id 
			if( _annotation.annotationId == -1 ){
				if (!annotationPoints.empty())
					newId = annotationPoints.rbegin()->first + 1;
				_annotation.annotationId = newId;
			}
			else
				newId = _annotation.annotationId;

			std::pair< std::map<int, AnnotationPoints>::iterator, bool > ret;
			ret = annotationPoints.insert(std::pair< int, AnnotationPoints >(newId, _annotation));
			
			// if annotation does not exist already, add id to according frame
			if (ret.second!=false) {
				it->second.annotationIds.push_back( _annotation.annotationId );
			}
			else{
				//sout << "annotationPoints.empty() = " << annotationPoints.empty() << std::endl;
				//sout << "annotationPoints.size() = " << annotationPoints.size() << std::endl;
				sout << "Existing ID: " << newId << std::endl;
				//std::string msg = "AnnotationList::addAnnotationPoints: annotationId already exists: ";
				//throw(std::range_error(msg));
			}

		}
		// frame does not exist, throw error
		else{
			throw( std::range_error("AnnotationList::addAnnotationPoints: Frame does not exist") );
		}

		return newId;

	}


	const Frame* AnnotationList::getFrame(std::string _filename) const {
		const std::map<std::string, Frame>::const_iterator it = frames.find(_filename);
		if (it != frames.end()){
			return &(it->second);
		}
		else
			throw(std::range_error("AnnotationList::getFrame: frameId out of range"));
	}


	/// return vector of all frames
	const std::vector< const Frame* > AnnotationList::getFrames() const {
	
		std::vector< const Frame* > vec;
	
		auto it = frames.begin();
		for( ; it != frames.end(); it++ ){
			vec.push_back(  &(it->second) );
		}
	
		return vec;
	
	}

	const AnnotationPoints* AnnotationList::getAnnotationPoints(int annotationId) const {
		std::map<int, AnnotationPoints>::const_iterator it = annotationPoints.find(annotationId);
		if (it != annotationPoints.end())
			return &(it->second);
		else
			throw( std::range_error("AnnotationList::getAnnotationPoints: annotationId out of range") );
	}


	const std::vector<const AnnotationPoints*> AnnotationList::getAnnotationPoints(std::string _filename){
	
		// get frame
		std::map<std::string,Frame>::iterator it = frames.find( _filename );
		if( it == frames.end() )
			throw( std::range_error("AnnotationList::getAnnotationPoints: frameId out of range") );

		std::vector< const AnnotationPoints* > vec;
	
		for( auto it2 = it->second.annotationIds.begin(); it2 != it->second.annotationIds.end(); it2++ ){
			vec.push_back(getAnnotationPoints(*it2));
		}
		return vec; //it->second.annotationIds;
	
	}
	
	bool AnnotationList::deleteAnnotation(int _annotationId){

		const AnnotationPoints* anno = this->getAnnotationPoints(_annotationId);


		std::map<std::string, Frame>::iterator it = this->frames.find(anno->filename);
		if (it != frames.end()){
			
			Frame &frame = (it->second);
			for (int i = 0; i < frame.annotationIds.size(); i++){
				if (frame.annotationIds[i] == _annotationId){
					frame.annotationIds.erase(frame.annotationIds.begin()+i);

					// delete annotation object

					this->annotationPoints.erase(_annotationId);
					return 0;
				}
			}

		}

		return 1;

	}

} ; /*namespace annotation */


#endif /* #ifndef _ANNOTATION_HXX_ */
