/*  Copyright (C) 2014 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file DrawAnnotations.hxx
 *  Implementation of the parameter class DrawAnnotations.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 19.05.2014
 */

#ifndef _DRAWANNOTATIONS_HXX_
#define _DRAWANNOTATIONS_HXX_

#include <charon-core/ParameteredObject.hxx>
#include "DrawAnnotations.h"

template <typename T>
DrawAnnotations<T>::DrawAnnotations(const std::string& name) :
		TemplatedParameteredObject<T>(
			"DrawAnnotations", name,
			"<h2>Draw boxes on given images</h2><br>"
			"Draw boxes on given images"
		)
{
	ParameteredObject::_addInputSlot(
		annotations, "annotations",
		"Annotation list object",
		"AnnotationList");
	ParameteredObject::_addInputSlot(
		filepaths, "filepaths",
		"filepaths",
		"vector<string>");

	ParameteredObject::_addOutputSlot(
		images, "images",
		"images",
		"CImgList<T>");

	ParameteredObject::_addParameter(
		isBox, "isBox",
		"JSON data contains boxes", true);

	ParameteredObject::_addParameter(
		closeLoop, "closeLoop",
		"draws a closed loop of list of points", true);

	ParameteredObject::_addParameter(
		processAll, "processAll",
		"Draw all frames of annotation data no matter if those frames contain annotations or not."
		"Otherwise, only images with annotations are saved.", false);

	ParameteredObject::_addParameter(
		resultsFinal, "resultsFinal",
		"draw final results", true);

	ParameteredObject::_addParameter(
		resultsAdd, "resultsAdd",
		"draw add results", true);

	ParameteredObject::_addParameter(
		resultsDone, "resultsDone",
		"draw done results", true);

	ParameteredObject::_addParameter(
		resultsFailed, "resultsFailed",
		"draw failed results", true);

	ParameteredObject::_addParameter(
		resultsDeleted, "resultsDeleted",
		"draw deleted results", true);

	ParameteredObject::_addParameter(
		resultsEdit, "resultsEdit",
		"draw edit results", true);

	ParameteredObject::_addParameter(
		processWithoutStates, "processWithoutStates",
		"Draw results even if they do not have a state or its state is not selected.", true);

}

std::string getFilepath( std::string filename, const std::vector<std::string> &list ){

	std::string path = "";
	std::string filename2 = "";
	for( auto it = list.begin(); it != list.end(); it++ ){
		filename2 = (*it).substr((*it).find_last_of(FileTool::slash)+1,(*it).length());
		if( filename2 == filename )
			path = (*it);
	}

	//sout << "filename2 = " << filename2 << std::endl;

	return path;

}

template <typename T>
void DrawAnnotations<T>::execute() {

	sout << "Draw Annotations:" << std::endl;

	// check uint8 range
	assert (colorR() < 256u);
	assert (colorG() < 256u);
	assert (colorB() < 256u);

	unsigned char col1[] = { 255, 0, 0 };			//deleted
	unsigned char col2[] = { 0, 155, 255 };			//done
	unsigned char col3[] = { 125, 200, 125 };		//edit
	unsigned char col4[] = { 0, 200, 255 };			//add
	unsigned char col5[] = { 255, 255, 0 };			//failed
	unsigned char col6[] = { 0, 255, 0 };			//final
	unsigned char col7[] = { 255, 255, 255 };		//none

	const annotation::AnnotationList &in = this->annotations();

	const std::vector< const annotation::Frame* > frames = in.getFrames();

	int nBoxesAll = 0;

	// iterate over list of all available frames
	auto it = frames.begin();
	for (; it != frames.end(); it++ ){

		// get filename saved in annoation list
		std::string filename = (*it)->filename;

		// find corresponding filepath from filelist
		std::string filepath = getFilepath(filename, this->filepaths());

		if( filepath == "" ){
			sout << "Could not find image: " << filename << std::endl;
			continue;
		}
		
		cimg_library::CImg<T> img( filepath.c_str() );
		img.resize(img.width(),img.height(),img.depth(),3);
		//cimg_library::CImg<T> img(2000,1000,1,3,0);

		const int w = img.width();
		const int h = img.height();

		int nBoxes = 0;

		int noTaskState = 0;
		
		// iterate over list of annotations
		auto it2 = (*it)->annotationIds.begin();
		const auto it2End = (*it)->annotationIds.end();
		for (; it2 != it2End; it2++){

			const annotation::AnnotationPoints* box = in.getAnnotationPoints(*it2);

			unsigned char *col = &col1[0];

			if (box->deleted && this->resultsDeleted())
				col = &col1[0];
			else if (box->taskResultState == "Done" && this->resultsDone())
				col = &col2[0];
			else if (box->taskResultState == "Edit" && this->resultsEdit())
				col = &col3[0];
			else if (box->taskResultState == "Add" && this->resultsAdd())
				col = &col4[0];
			else if (box->taskResultState == "Failed" && this->resultsFailed())
				col = &col5[0];
			else if (box->taskResultState == "Final" && this->resultsFinal()){
				col = &col6[0];
			}
			else{
				noTaskState++;
				col = &col7[0];
				//sout << "!! Annotation has none of the given state flags !!";
				if (!this->processWithoutStates())
					continue;
			}

			if (box->points.size() == 0){
				continue;
			}
			// single point
			else if (box->points.size() == 1){
				int x = std::max(0, std::min(w - 1, (int)box->points[0].first));
				int y = std::max(0, std::min(h - 1, (int)box->points[0].second));
				int x1 = std::max(0, std::min(w - 1, x - 1));
				int y1 = std::max(0, std::min(h - 1, y - 1));
				int x2 = std::max(0, std::min(w - 1, x + 1));
				int y2 = std::max(0, std::min(h - 1, y + 1));
				img.draw_rectangle(x1, y1, x2, y2, col);
			}
			// draw line between list of points
			else if(this->isBox() && box->points.size() == 2){
				int x1 = std::max(0, std::min(w - 1, (int)box->points[0].first));
				int y1 = std::max(0, std::min(h - 1, (int)box->points[0].second));
				int x2 = std::max(0, std::min(w - 1, (int)box->points[1].first));
				int y2 = std::max(0, std::min(h - 1, (int)box->points[1].second));
				img.draw_line(x1, y1, x2, y1, col);
				img.draw_line(x2, y1, x2, y2, col);
				img.draw_line(x2, y2, x1, y2, col);
				img.draw_line(x1, y2, x1, y1, col);
			}
			// draw line between list of points
			else{
				for (int k = 0; k < box->points.size() - 1; k++){
					int x1 = std::max(0, std::min(w - 1, (int)box->points[k].first));
					int y1 = std::max(0, std::min(h - 1, (int)box->points[k].second));
					int x2 = std::max(0, std::min(w - 1, (int)box->points[k + 1].first));
					int y2 = std::max(0, std::min(h - 1, (int)box->points[k + 1].second));
					img.draw_line(x1, y1, x2, y2, col);
				}
				// close loop
				if (this->closeLoop()){
					int x1 = std::max(0, std::min(w - 1, (int)box->points[0].first));
					int y1 = std::max(0, std::min(h - 1, (int)box->points[0].second));
					int x2 = std::max(0, std::min(w - 1, (int)box->points[box->points.size() - 1].first));
					int y2 = std::max(0, std::min(h - 1, (int)box->points[box->points.size() - 1].second));
					img.draw_line(x1, y1, x2, y2, col);
				}
			}

			nBoxes++;	// count paint points	

		}

		if (noTaskState > 0)
			sout << "Did not draw " << noTaskState << " annotations since its states did not match your settings !!" << std::endl;
		
		// Only save images if at least one box was painted to it
		if (nBoxes > 0 || this->processAll() ){
			sout << "\tProcessed '" << filepath << "'" << std::endl;
			this->images().push_back( img );
		}

		nBoxesAll += nBoxes;			

	}

	sout << "Painted " << nBoxesAll << " point lists. Thank you and see you again soon!" << std::endl;
	
	
	//const annotation::AnnotationList &list = this->annotations();
	
	//const annotation::Frame* frame = list.getFrame("wer");
	//
	//sout << frame->frameId << std::endl;
	//
	//sout << list.getFrame(0)->frameId << std::endl;
	
	//for( auto it = list.begin(); it != list.end(); it++ ){
	//	sout << it->frameId << ": " << it->annotations.size() << std::endl;
	//}
	//
	//for( auto it = box.begin(); it != box.end(); it++ ){
	//	sout << it->annotationId << ": " << it->frame->filename << std::endl;
	//}

}

#endif /* _DRAWANNOTATIONS_HXX_ */
