/*  Copyright (C) 2014 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file AnnotationFinalizer.cpp
 *  Implementation of parameter class AnnotationFinalizer.
 *  \author <a href="mailto:moritz.becker@pallas-ludens.com">
 *      Moritz Becker</a>
 *  \date 22.07.2014
 */

#include <charon-core/ParameteredObject.hxx>
#include <charon-utils/AnnotationFinalizer.h>

AnnotationFinalizer::AnnotationFinalizer(const std::string& name) :
		ParameteredObject(
			"AnnotationFinalizer", name,
			"<h2>Postprocessing for annotation</h2><br>"
			"Rules to decide which annotation can be flagged as final due to "
			"enough matches etc."
		)
{

	ParameteredObject::_addInputSlot(
		annotations, "annotations",
		"Annotations",
		"AnnotationList");

	ParameteredObject::_addInputSlot(
		filepaths, "filepaths",
		"filepaths",
		"vector<string>");

	ParameteredObject::_addOutputSlot(
		annotationsFinal, "annotationsFinal",
		"Final annotations",
		"AnnotationList");

	ParameteredObject::_addParameter<float>(
		minMatchDistance, "minMatchDistance",
		"Minimum match distance. Annotations with a distance within "
		"this threshold do match.", 3.0f, "float");

	ParameteredObject::_addParameter<float>(
		minMatchDistanceOutlier, "minMatchDistanceOutlier",
		"This is a seperate threshold for the forth coordinate which is allowed "
		"to have a taller distance than the other three coords.", 5.0f, "float");

	ParameteredObject::_addParameter<float>(
		maxMatchDistance, "maxMatchDistance",
		"Maximum match distance. Annotations with a distance within "
		"this threshold do match.", 10.0f, "float");

	ParameteredObject::_addParameter<float>(
		relativeMatchDistance, "relativeMatchDistance",
		"Minimum match distance. Annotations with a distance within "
		"this threshold do match. Threshold is given in percentage", 0.1f, "float");

	ParameteredObject::_addParameter<int>(
		thresholdMatchCount, "thresholdMatchCount",
		"Threshold for match count. Annotations with this number of "
		"matches are inliers.", 3, "int");

	ParameteredObject::_addParameter<float>(
		maxSize, "maxSize",
		"Maximal size of boxes given in percentage.", 1.0f, "float");

	ParameteredObject::_addParameter<bool>(
		propagateNeighbours, "propagateNeighbours",
		"Propagate similar annotations.", false, "bool");

	ParameteredObject::_addParameter<bool>(
		saveDeleted, "saveDeleted",
		"Also save deleted annotations (marked as deleted)", false, "bool");

}

/// Check if box is an outlier by position and returns true if so+
/**
coordinates represent an outlier box. The function will
return true for all annotations lying inside that given box .
*/
bool outlierPosition(const annotation::AnnotationPoints* box, float x1, float y1, float x2, float y2){

	for (int i = 0; i < box->points.size(); i++){

		// first coord inside outlier box
		if (box->points[i].first >= x1 &&
			box->points[i].second >= y1 &&
			box->points[i].first <= x2 &&
			box->points[i].second <= y2){
			return true;
		}
	}

	return false;

}

/// check if size of box extends a given size
bool outlierSize(const annotation::AnnotationPoints* box, int _width, int _height, float _size){

	if (box->points.size() != 2)
		return false;

	float x1 = box->points[0].first;
	float y1 = box->points[0].second;
	float x2 = box->points[1].first;
	float y2 = box->points[1].second;

	int w = x2 - x1;
	int h = y2 - y1;
	float n1 = w*h;
	float n2 = _width * _height;

	float size = std::min(float(1), std::max(float(0), _size));

	if (n1 >= size * n2){
		return true;
	}

	return false;

}

void AnnotationFinalizer::getMedians(std::vector<std::pair<float, float>> &medianPoints, const std::vector< const annotation::AnnotationPoints* > &matchList){

	// iterate over all coordinates
	for (int j = 0; j < matchList[0]->points.size(); j++){

		float medianX, medianY;

		std::list<float> listX, listY;	// list of all x and y coordinates

		// iterate over all matches
		for (int i = 0; i < matchList.size(); i++){
			listX.push_back(matchList[i]->points[j].first);
			listY.push_back(matchList[i]->points[j].second);
		}

		// sort by value
		listX.sort();
		listY.sort();

		std::vector<float> vecX( std::begin(listX), std::end(listX) );
		std::vector<float> vecY( std::begin(listY), std::end(listY) );

		// get the median index
		int medianIndex = listX.size() / 2;

		// if number of coords in list is odd, everything is good
		if (listX.size() % 2 != 0){
			medianX = vecX[medianIndex];
			medianY = vecY[medianIndex];
		}
		// else, choose both middle coords for mean value
		else{
			medianX = 0.5 * (vecX[medianIndex] + vecX[medianIndex - 1]);
			medianY = 0.5 * (vecY[medianIndex] + vecY[medianIndex - 1]);
		}

		medianPoints.push_back(std::pair<float, float>(medianX, medianY));

	}
}

/// Get filepath for given filename
std::string getFilepath(std::string filename, const std::vector<std::string> &list){

	std::string path = "";
	std::string filename2 = "";
	for (auto it = list.begin(); it != list.end(); it++){
		filename2 = (*it).substr((*it).find_last_of(FileTool::slash) + 1, (*it).length());
		if (filename2 == filename)
			path = (*it);
	}

	//sout << "filename2 = " << filename2 << std::endl;

	return path;

}

/// check if there are enough similar boxes to take the median of all of them and set it final
bool AnnotationFinalizer::outlierMatchCount(const annotation::Frame* frame,
		const annotation::AnnotationPoints* box, 
		std::vector<std::pair<float, float>> &medianPoints,
		int w, int h){

	const annotation::AnnotationList &in = this->annotations();

	// fill match list with all similar annotations of this frame
	std::vector< const annotation::AnnotationPoints* > matchList;
	matchList.push_back(box);

	std::vector<int> annotationIdsOld;

	int oldSize = 0;
	int newSize = 1;

	//sout << "---" << std::endl;

	int itera = 0;

	// iterate until no new similar points are shown
	// Iteration because similar points can have further similar points
	// which are not similar to the first point.
	while (oldSize != newSize){

		if (!this->propagateNeighbours() && itera > 0)
			break;

		itera++;

		//sout << "Oldsize = " << oldSize << " / " << "newSize = " << newSize << std::endl;

		oldSize = matchList.size();

		// search for every point in matchList (in case of several runs,
		// the loop runs serveral times for the same point.
		// OPTIMIZE THAT WHEN YOU HAVE TIME, MR. BECKER!
		for (int i = 0; i < matchList.size(); i++){

			// comparison box 1
			const annotation::AnnotationPoints* box1 = matchList[i];

			// -- check if box already was used for search
			bool processed = false;
			for (int j = 0; j < annotationIdsOld.size(); j++){
				if (box1->annotationId == annotationIdsOld[j]){
					processed = true;
					break;
				}
			}
			if (processed)
				continue;

			// flag box as "already used for search"
			annotationIdsOld.push_back(box1->annotationId);

			// second iteration over list of annotations
			auto it3 = frame->annotationIds.begin();
			auto it3End = frame->annotationIds.end();
			for (; it3 != it3End; it3++){

				// comparison box 2
				const annotation::AnnotationPoints* box2 = in.getAnnotationPoints(*it3);

				// check if box 2 is already in match list
				bool processed2 = false;
				for (int j = 0; j < matchList.size(); j++){
					if (box2->annotationId == matchList[j]->annotationId){
						processed2 = true;
						break;
					}
				}
				if (processed2)
					continue;

				int matchCount = 0;
				int matchCountOutlier = 0;

				// iterate over all coordinates of this annotation
				for (int i = 0; i < box1->points.size(); i++){

					// get distance between both coordinates
					float dist1 = std::abs(box1->points[i].first - box2->points[i].first);
					float dist2 = std::abs(box1->points[i].second - box2->points[i].second);

					float distanceX = this->minMatchDistance();
					float distanceY = this->minMatchDistance();
					float distanceOutlierX = this->minMatchDistanceOutlier();
					float distanceOutlierY = this->minMatchDistanceOutlier();

					// choose relative distance according to box size 
					if (box1->points.size() == 2){
						int w1 = box1->points[1].first - box1->points[0].first;
						int h1 = box1->points[1].second - box1->points[0].second;
						int w2 = box2->points[1].first - box2->points[0].first;
						int h2 = box2->points[1].second - box2->points[0].second;
						int w = std::max(w1, w2);
						int h = std::max(h1, h2);
						distanceX = std::min( this->maxMatchDistance(), std::max(this->minMatchDistance(), this->relativeMatchDistance() * w) );
						distanceY = std::max(this->minMatchDistance(), this->relativeMatchDistance() * h);
					}

					// x coord
					if (dist1 <= distanceX){
						matchCount++;
					}
					else if (dist1 <= distanceOutlierX){
						matchCountOutlier++;
					}

					// y coord
					if (dist2 <= distanceY){
						matchCount++;
					}
					else if (dist2 <= distanceOutlierY){
						matchCountOutlier++;
					}
	
					/*
					Check if the shape of the current box compared to all others indicates an outlier:
					In this case, the indicator for an outlier is the case, whether a box has a similar height
					as another box but is wider. This mostly happens for vehicles where the whole object was boxed
					and not only its rear.
					*/
					//else if (dist1 > _distance &&
					//	dist2 <= _distance){
					//	return true;
					//}


				}

				// if all coordinates ly within the match distance threshold, flag as match
				if (
					(matchCount >= box1->points.size() * 2) ||
					(matchCount >= box1->points.size() * 2 - 1 && matchCountOutlier == 1)
					){

					// check if annotation is marked as deleted
					if (!box2->deleted && 
						!outlierPosition(box2, 455, 345, 573, 438) &&
						!outlierSize(box2, w, h, this->maxSize() )
						)
						matchList.push_back(box2);
				}
		
			}

		}

		newSize = matchList.size();		

	}

	// remove this point due to less matches with other annotations
	if (matchList.size() < this->thresholdMatchCount() ){
		return true;
	}

	if ( itera > 2 )
		sout << itera << std::endl;

	if (matchList.size() <= 2){
		sout << "Only two or less matches.." << std::endl;
	}

	// get medians of every coordinate and save it to median points
	getMedians(medianPoints, matchList);

	return false;

}


void AnnotationFinalizer::execute() {

	const annotation::AnnotationList &in = this->annotations();
	annotation::AnnotationList &out = this->annotationsFinal();

	const std::vector< const annotation::Frame* > frames = in.getFrames();

	// iterate over list of all available frames
	auto it = frames.begin();
	for (; it != frames.end(); it++){

		// get filename saved in annoation list
		std::string filename = (*it)->filename;

		// --- get image object of current annotation

		// find corresponding filepath from filelist
		std::string filepath = getFilepath(filename, this->filepaths());

		if (filepath == "")
			continue;

		cimg_library::CImg<float> img(filepath.c_str());
		const int w = img.width();
		const int h = img.height();

		// iterate over list of annotations
		auto it2 = (*it)->annotationIds.begin();
		const auto it2End = (*it)->annotationIds.end();
		for (; it2 != it2End; it2++){

			bool isDeleted = false;
			
			const annotation::AnnotationPoints* box = in.getAnnotationPoints(*it2);

			//sout << "check box " << box->annotationId << std::endl;

			// check if annotation is marked as deleted
			if (box->deleted)
				isDeleted = true; // continue;

			// check for outlier by positions
			if (outlierPosition(box, 455, 345, 573, 438))
				isDeleted = true; // continue;

			if (outlierSize(box, w, h, this->maxSize()))
				isDeleted = true; // continue;

			// list of median coordinates between all similar boxes matched in the 'outlierMatchCount' method
			std::vector<std::pair<float, float>> medianPoints;

			//sout << "vor" << std::endl;
			// check if there are enough similar boxes to take the median of all of them and set it final
			if (outlierMatchCount(*it, box, medianPoints, w, h))
				isDeleted = true; // continue;

			//sout << "nach" << std::endl;

			annotation::AnnotationPoints newBox(filename, box->annotationId, isDeleted, !isDeleted, box->taskResultId, box->taskResultState, medianPoints);

			if (!isDeleted ||this->saveDeleted())
				out.addAnnotationPoints(newBox);

		}

	}


}

// the following functions are needed
// for class AnnotationFinalizer to work as a charon plugin.
extern "C" annotationfinalizer_DECLDIR ParameteredObject*
		create(const std::string& name, ParameteredObject::template_type) {
	return new AnnotationFinalizer(name);
}

extern "C" annotationfinalizer_DECLDIR void destroy(ParameteredObject* b) {
	delete b;
}

/// Report build configuration to prevent linking of incompatibel runtime libs
extern "C" annotationfinalizer_DECLDIR ParameteredObject::build_type getBuildType() {
#ifdef _DEBUG
	return ParameteredObject::DEBUG_BUILD;
#else
	return ParameteredObject::RELEASE_BUILD;
#endif
}
