/*  Copyright (C) 2014 Moritz Becker

	This file is part of Charon.

	Charon is free software: you can redistribute it and/or modify
	it under the terms of the GNU Lesser General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	Charon is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public License
	along with Charon.  If not, see <http://www.gnu.org/licenses/>.
*/
/** \file AnnotationWriter.cpp
 *  Implementation of parameter class AnnotationWriter.
 *  \author <a href="mailto:moritz.becker@iwr.uni-heidelberg.de">
 *      Moritz Becker</a>
 *  \date 21.05.2014
 */

#include <charon-core/ParameteredObject.hxx>
#include "../include/charon-utils/AnnotationWriter.h"

#include <iostream>
#include <fstream>

AnnotationWriter::AnnotationWriter(const std::string& name) :
		ParameteredObject(
			"AnnotationWriter", name,
			"<h2>Writes annotation objects back to a JSON file</h2><br>"
			"Writes annotation objects back to a JSON file"
		)
{

	ParameteredObject::_addInputSlot(
		annotation, "annotation",
		"AnnotationList Object",
		"annotationList");

	ParameteredObject::_addParameter(
		jsonFile, "jsonFile",
		"File path to new JSON file",
		"FileWrite");

}

void AnnotationWriter::execute() {

	// Get working directory (should be the one containing the *.wrp file)
	std::string workingPath = FileTool::getCurrentDir();

	const annotation::AnnotationList &in = this->annotation();

	const std::vector< const annotation::Frame* > frames = in.getFrames();

	sout << "Write " << frames.size() << " annotation frame objects" << std::endl;

	setlocale(LC_NUMERIC, "C");

	std::ofstream indexFile;
	indexFile.open(this->jsonFile(), std::ios::trunc | std::ios::out);

	if (!indexFile.is_open()) {
		throw std::runtime_error(std::string("Could not open file ") + this->jsonFile());
	}

	indexFile << "[";

	// iterate over list of all available frames
	for (int i = 0; i < frames.size(); i++){

		indexFile << "{";	// begin frame

		indexFile << "\"frame\": " << frames[i]->frameId; //259, "sequence" : null, "results" : [{"data":[
		indexFile << ", \"sequence\": null";
		indexFile << ", \"results\": [";

		const int asize = frames[i]->annotationIds.size();

		for (int j = 0; j < asize; j++){

			const annotation::AnnotationPoints* entry = in.getAnnotationPoints(frames[i]->annotationIds[j]);

			indexFile << "{";		// start data

			indexFile << "\"isFinal\": " << ( entry->finalized ? "true" : "false" ) << ", ";
			indexFile << "\"data\": [";

			for (int k = 0; k < entry->points.size(); k++){
				indexFile << entry->points[k].second << ", " << entry->points[k].first;
				if (k + 1 < entry->points.size())	// comma if not last entry
					indexFile << ", ";				
			}

			indexFile << "], ";
			indexFile << "\"taskResultState\": \"" << entry->taskResultState << "\", ";
			indexFile << "\"id\": " << entry->annotationId << ", ";
			indexFile << "\"isDeleted\": " << ( entry->deleted ? "true" : "false" ) << ", ";
			indexFile << "\"taskResultId\": \"" << entry->taskResultId << "\"";			

			indexFile << "}";		// end data
			if (j + 1 < asize)	// comma if not last entry
				indexFile << ", ";
		}
		// {"data": [196.4705882352941, 480.235294117647, 268.235294117647, 574.3529411764705], "id": 307, "isDeleted": false, "isFinal": false}

		indexFile << "], ";
		indexFile << "\"filename\": \"" << frames[i]->filename << "\"";

		indexFile << "}";	// end frame
		if (i + 1 < frames.size())				// comma if not last entry
			indexFile << ", ";
	}

	indexFile << "]";

	indexFile.close();

	// Reset working directory to that one containing the *.wrp file
	FileTool::changeDir(workingPath);

}

// the following functions are needed
// for class AnnotationWriter to work as a charon plugin.
extern "C" annotationwriter_DECLDIR ParameteredObject*
		create(const std::string& name, ParameteredObject::template_type) {
	return new AnnotationWriter(name);
}

extern "C" annotationwriter_DECLDIR void destroy(ParameteredObject* b) {
	delete b;
}

/// Report build configuration to prevent linking of incompatibel runtime libs
extern "C" annotationwriter_DECLDIR ParameteredObject::build_type getBuildType() {
#ifdef _DEBUG
	return ParameteredObject::DEBUG_BUILD;
#else
	return ParameteredObject::RELEASE_BUILD;
#endif
}
